#!/usr/bin/env python
from sympy import Matrix, symbols, Symbol, init_printing, Function, dsolve, checkodesol, sin, cos, pi, Eq, \
    solveset, collect, cancel, simplify, lambdify, acos, sqrt, I, integrate, factor, factor_list, Matrix, \
    expand, init_printing, init_session, sympify
import re
from IPython.display import display
from sympy import Add, Mul, Symbol, Equality, Derivative, shape
import sys


class systemEqns:
    def __init__(self):
        self.ns = self.nextSymbol()
        self.symbols_used = []
        self.w = Symbol('omega')
        self.fnDict = {}

    def nextSymbol(self, prefix = 'a'):
        idx = 1
        while True:
            sym = Symbol("{}_{}".format(prefix, idx))
            yield sym
            idx+=1

    def getFunction(self, fname):
        if fname not in self.fnDict:
            a,b = ( self.ns.__next__(), self.ns.__next__() )
            self.symbols_used.extend((a,b))
            self.fnDict[fname] = ( a, b)
        return self.fnDict[fname]

    def recurseExpr(self, expr):
        if isinstance(expr, Derivative):
            fn = expr.args[0]
            assert(isinstance(fn, Function))
            fnName = fn.__class__.__name__
            r0, r1 = self.getFunction(fnName)
            return (r1*self.w, -r0*self.w)

        elif isinstance(expr, Add):
            terms = []
            for term in expr.args:
                terms.extend( self.recurseExpr(term) )
            return terms
        elif isinstance(expr, Mul):
            product = sympify(1)
            funcName = None
            trigFn = None
            for fact in expr.args:
                if fact.is_Number:
                    # If integer or rational number
                    product = product * fact
                elif fact.is_Function:
                    # If is a sympy function
                    if type(fact) in [ sin, cos]:
                        trigFn = sin
                    else:
                        funcName = fact.__class__.__name__
                elif fact.is_Pow:
                    # Then it is probably something like 1/C and we assume there is no function in
                    # the denominator (because we assuming this is a linear differential equation).
                    product = product * fact

                elif fact.is_Symbol:
                    # This captures expression like V_{in} that don't otherwise
                    # register as functions but we also want to treat other unspecified symbols
                    # as symbolic constants
                    fName = str(fact)
                    if fName.startswith("V_"):
                        funcName = fName
                    else:
                        product = product * fact

            #assert funcName is not None
            # If funcName is None then this is no longer a linear equation but one with a constant term added
            r0, r1 = self.getFunction(funcName)
            #
            ## If there is a sin or cos term in there then that means that it only contributes to one
            # of the linear equations.  Either just the sin expression or the cos expression (as sin and
            # cos are orthogonal to each other).
            # 
            if trigFn == sin:
                return [(0, r1*product)]
            elif trigFn == cos:
                return [(r0*product, 0)]
            else:
                return [(r0*product, r1*product)]


    def extractCoeffsFromDiffEqn(self, expr):
        """ This function is limited to linear equations of functions 
            of the form:
            c1*y' = a1*y1 + a2*y2 + ... +an*yn
            where a1...an are numbers.
            
            It takes a linear equation as the `expr` and tries solving it with
            sines and cosine coefficients in the form (a1, a2), (a3 a4), ...
            where the order of the coefficients could be in a different order.
            
            Each equation I am fitting with an*sin(wt) + an+1 * cos(wt) but I am wondering
            if it would have been normal to have the cos term first.
            
            """


        if isinstance(expr, Equality):
            assert(isinstance(expr.lhs, Derivative))  # If an equality expect the lhs to be a derivative
            assert(isinstance(expr.lhs.args[0], Function))  # Assume that the Derivative is of a function

            # We have to swap positions because under differentiation a sine becomes a cosine and vice versa
            resList = [self.recurseExpr(expr.lhs)]
            
            
            remaining = self.recurseExpr(expand(expr.rhs))
            # Reverse the sign of everything on the right hand side  to 
            # move it to the left hand side because
            # we want the whole equation to equal zero on the right hand side.
            for a,b in remaining:
                resList.append((-a, -b))
            return resList

    def getSymIdx(self, term):
        result = None
        for sym in self.symbols_used:
            # do the term term
            
            if sym in term.free_symbols:
                if not term.args:
                    thisList[sortedSyms.index(sym)] = 1
                   
                elif isinstance(term, Mul):
                    factors = sympify(1)
                    for f in term.args:
                        if f not in self.symbols_used:
                            factors = factors * f
                        else:
                            thisSym = f
                    idx = self.symbols_used.index(sym)
                    result = (idx, factors, sym)
                    break
        return result


    def orderedMatrix(self, results):
        from operator import itemgetter
        mList = []
        lsyms = len(self.symbols_used)
        for res in results:
            l1 = [ sympify(0)]  * lsyms
            l2 = [ sympify(0)]  * lsyms
            for x1, x2 in res:
               
                #print (x1, x2)
                if x1 != 0:
                    idx1, factors1, sym1 = self.getSymIdx(x1)
                    #print (idx1, factors1, sym1 )
                    l1[idx1] += factors1
                if x2 != 0:
                    idx2, factors2, sym2 = self.getSymIdx(x2)
                    #print (idx2, factors2, sym2 )
                    l2[idx2] += factors2
            mList.append(l2)
            mList.append(l1)

        #print (mList)
        
        # At this point I am trying to diagonlise the matrix with respect to w (omega) by swapping columns
        # and also trying to keep track of the co-efficients from a0, a1, .... an
        
        rows, cols = 2*len(results), lsyms 
        mtmp = []
        wcols = [] # columns with w
        for colidx in range(cols):
            for rowidx in range(rows):
                if self.w in mList[rowidx][colidx].free_symbols:
                    wcols.append ((colidx, rowidx))
                    break
            else:
                wcols.append( (colidx, 999) )


        wcols = sorted(wcols, key=itemgetter(1))
        
        reordered_syms = []
        for colidx, rowidx in wcols:
            mtmp.append ( [ row[colidx] for row in mList] )
            reordered_syms.append (self.symbols_used[colidx] )

        M = Matrix(mtmp).T


        #print(mList)

        return M, reordered_syms



if __name__ == "__main__":
    A_0, A_1, A_2, B_1, B_2, R1, R2, R3, Vin, C1, C2 = symbols("A_0 A_1 A_2 B_1 B_2 R_1 R_2 R_3 V_{in} C_1 C_2")
    Vc1 = Function('Vc1')
    Vc2 = Function('Vc2')
    Vc3 = Function('Vc3')
    R1,R2, Vin, Vc1, Vc2, C1,C2
    Vin, Vc, C, R, i, t, f, A, tau, w = symbols("V_{in} V_c C R i t f A tau omega")

    m = Matrix([[R1, -R1], [-R1, R1-R2]])
    v = Matrix([Vin-Vc1(t), Vc2(t)])
    mtmp = m.inv()*v


    eqn_1 = Eq(C1 * Vc1(t).diff()/C, mtmp[0]/C)
    eqn1 = eqn_1.subs([(R1, R), (R2, R), (C1,C)])

    eqn_2 = Eq(Vc2(t).diff(), simplify(mtmp[1]/C))
    eqn2 = eqn_2.subs([(R1, R), (R2, R)])

    display(eqn1)
    display(eqn2)

    sysEqns = systemEqns()

    res1 = sysEqns.extractCoeffsFromDiffEqn(eqn1)
    print (res1)
    res2 = sysEqns.extractCoeffsFromDiffEqn(eqn2)
    print (res2)

    sortedSyms = sysEqns.symbols_used
    print (sortedSyms)

    M1, _ = sysEqns.orderedMatrix([res1, res2])
    display(M1)
    
    
    # Next test
    eqn1 = Eq(Derivative(Vc1(t), t), (A*sin(w*t) - Vc1(t) - Vc3(t))/(C*R))
    eqn2 = Eq(Derivative(Vc2(t), t), -Vc3(t)/(C*R))
    eqn3 = Eq(Derivative(Vc3(t), t), (-A*sin(w*t) + Vc1(t) - Vc2(t) + Vc3(t))/(C*R))
    
    display(eqn1)
    display(eqn2)
    display(eqn3)
    
    sysEqns = systemEqns()

    res1 = sysEqns.extractCoeffsFromDiffEqn(eqn1)
    print (res1)
    res2 = sysEqns.extractCoeffsFromDiffEqn(eqn2)
    print (res2)
    res3 = sysEqns.extractCoeffsFromDiffEqn(eqn3)
    print (res3)
    M1, syms_used = sysEqns.orderedMatrix([res1, res2, res3])
    display(M1)    
    print (syms_used)
    
        
